import os
import json
import uuid
import threading
import subprocess
import glob
from enum import Enum
import logging


class Progress_status(Enum):
    IDLE = 'IDLE'
    EXTRACTING = 'EXTRACTING'
    UPLOADING = 'UPLOADING'
    UPLOAD_COMPLETE = 'UPLOAD_COMPLETE'
    FAILED = 'FAILED'


class SupportFileHandler:

    def __init__(self, message_handler):
        global logger
        logger = logging.getLogger('myur.daemon.' + __name__)
        self.message_handler = message_handler
        self.blob_uploader = message_handler.upload_to_blob_handler
        self.processStateFileName = "support-file-process-state.json"
        self.processStateFileNameFullPath = ""  # is set through create_support_file_process_state_file_if_not_exists
        self._create_support_file_process_state_file_if_not_exists()
        self.state = self._get_state_from_file()
        self._reset()

    def generate_support_file(self, tags):
        current_status = self._get_status()
        if current_status != Progress_status.IDLE.value and current_status != Progress_status.FAILED.value:
            return SupportFileHandlerResponse(self._get_status(), self._get_correlation_id(), 409,
                                              "Process is already running")

        self._update_correlation_id(self._generate_uuid())

        x = threading.Thread(target=self._start_process_flow, args=(tags,))
        x.daemon = True
        x.start()

        return SupportFileHandlerResponse(self._get_status(), self._get_correlation_id(), 200, "Extraction started")

    def delete_support_file(self):
        for file in glob.glob(self._get_support_file_directory_path() + "*.zip"):
            if os.path.exists(file):
                os.remove(file)

    def _start_process_flow(self, tags):
        try:
            self._update_status(Progress_status.EXTRACTING.value)
            self._send_status_update_to_hub()

            # self._simulate_support_file()
            subprocess.call(['sh', './generate_support_file.sh'])

            support_file_path = ""
            for file in glob.glob(self._get_support_file_directory_path() + "*.zip"):
                support_file_path = file
            self._update_file_size(os.path.getsize(support_file_path))

            new_file_name = self._get_support_file_directory_path() + self._get_correlation_id() + ".zip"
            self._rename_file(support_file_path, new_file_name)
            self._update_status(Progress_status.UPLOADING.value)
            self._send_status_update_to_hub()
            self.blob_uploader.upload_files("SupportLogFiles", new_file_name, tags, True)

            self._update_status(Progress_status.UPLOAD_COMPLETE.value)
            self._send_status_update_to_hub()

            self._reset()
        except Exception as e:
            self._update_status(Progress_status.FAILED.value)
            self._update_message(str(e))
            self._send_status_update_to_hub()
            self._reset()

    def _send_status_update_to_hub(self):
        self.message_handler.send_message_to_hub(
            '{ "type" : "SupportFileStatus", "state" : ' + json.dumps(self.state) + ' }')

    def _reset(self):
        self._update_status(Progress_status.IDLE.value)
        self._update_correlation_id("")
        self._update_file_size(0)
        self._update_message("")

    def _create_support_file_process_state_file_if_not_exists(self):
        path = self._get_state_file_path()
        if not os.path.exists(path):
            os.makedirs(path)

        self.processStateFileNameFullPath = path + self.processStateFileName

        if not os.path.exists(self.processStateFileNameFullPath):
            d = json.loads('{"status": "'+Progress_status.IDLE.value+'", "correlationId" : "", "supportFileSize" : 0, "message" : "" }')
            self._write_to_status_file(json.dumps(d), path + self.processStateFileName)

    def _write_to_status_file(self, data, filepath):
        with open(filepath, 'w') as f:
            f.write(data)

    def _get_state_from_file(self):
        with open(self.processStateFileNameFullPath, "r") as f:
            state = json.loads(f.read())
        return state

    def _get_support_file_directory_path(self):
        if os.path.exists("/root/ur-serial"):
            directory = "/programs/support_file/"
        else:
            directory = "/tmp/support_file/"

        if not os.path.exists(directory):
            os.makedirs(directory)

        return directory

    def _get_state_file_path(self):
        if os.path.exists("/root/ur-serial"):
            directory = "/data/myUR/"
        else:
            directory = "/tmp/"
        return directory

    def _get_correlation_id(self):
        return self.state["correlationId"]

    def _get_status(self):
        return self.state["status"]

    def _update_correlation_id(self, correlationId):
        self.state["correlationId"] = correlationId
        self._write_to_status_file(json.dumps(self.state), self.processStateFileNameFullPath)

    def _update_status(self, status):
        self.state["status"] = status
        self._write_to_status_file(json.dumps(self.state), self.processStateFileNameFullPath)

    def _update_message(self, message):
        self.state["message"] = message
        self._write_to_status_file(json.dumps(self.state), self.processStateFileNameFullPath)

    def _update_file_size(self, size):
        self.state["supportFileSize"] = size
        self._write_to_status_file(json.dumps(self.state), self.processStateFileNameFullPath)

    def _generate_uuid(self):
        return str(uuid.uuid4())

    def _rename_file(self, current_path, new_path):
        os.rename(current_path, new_path)

    def _simulate_support_file(self):
        filepath = self._get_support_file_directory_path() + "someName.zip"
        with open(filepath, 'w') as f:
            f.write("data")


class SupportFileError(Exception):
    def __init__(self, message):
        self.message = message


class SupportFileHandlerResponse:
    def __init__(self, current_status, correlation_id, status_code, message):
        self.currentStatus = current_status
        self.correlationId = correlation_id
        self.statusCode = status_code
        self.message = message
